#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "config.h"
#include "imapfilter.h"
#include "pathnames.h"

#ifdef SSL_TLS
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/pem.h>


extern unsigned int options;
extern unsigned int flags;


int check_cert(X509 * pcert, unsigned char *pmd, unsigned int *pmdlen);
void print_cert(X509 * cert, unsigned char *md, unsigned int *mdlen);
int write_cert(X509 * cert);
int mismatch_cert(void);


/*
 * Get SSL/TLS certificate check it, maybe ask user about it and act
 * accordingly.
 */
int
get_cert(conn_t * conn)
{
	X509 *cert;
	unsigned char md[EVP_MAX_MD_SIZE];
	unsigned int mdlen;

	mdlen = 0;

	if (!(cert = SSL_get_peer_certificate(conn->ssl)))
		return ERROR_SSL;

	if (!(X509_digest(cert, EVP_md5(), md, &mdlen)))
		return ERROR_SSL;

	switch (check_cert(cert, md, &mdlen)) {
	case SSL_CERT_NONEXISTENT:
		print_cert(cert, md, &mdlen);
		if (flags & FLAG_DAEMON_MODE ||
		    write_cert(cert) == SSL_CERT_ACTION_REJECT)
			goto abort;
		break;
	case SSL_CERT_MISMATCH:
		print_cert(cert, md, &mdlen);
		if (flags & FLAG_DAEMON_MODE ||
		    mismatch_cert() == SSL_CERT_ACTION_ABORT)
			goto abort;
		break;
	case SSL_CERT_OK:
		if (options & OPTION_DETAILS_VERBOSE)
			print_cert(cert, md, &mdlen);
	}

	X509_free(cert);
	return 0;

abort:
	X509_free(cert);
	return ERROR_SSL;
}


/*
 * Check if the SSL/TLS certificate exists in the certificates file.
 */
int
check_cert(X509 * pcert, unsigned char *pmd, unsigned int *pmdlen)
{
	int r;
	FILE *fd;
	X509 *cert;
	unsigned char md[EVP_MAX_MD_SIZE];
	unsigned int mdlen;

	r = SSL_CERT_NONEXISTENT;
	cert = NULL;

	if (!exists_file(PATHNAME_CERT_FILE))
		return SSL_CERT_NONEXISTENT;

	fd = fopen(PATHNAME_CERT_FILE, "r");
	if (fd == NULL)
		return ERROR_FILE_OPEN;

	while ((cert = PEM_read_X509(fd, &cert, NULL, NULL)) != NULL) {
		if (X509_subject_name_cmp(cert, pcert) != 0 ||
		    X509_issuer_name_cmp(cert, pcert) != 0)
			continue;

		if (!X509_digest(cert, EVP_md5(), md, &mdlen) ||
		    *pmdlen != mdlen)
			continue;

		if (memcmp(pmd, md, mdlen) != 0) {
			r = SSL_CERT_MISMATCH;
			break;
		}
		r = SSL_CERT_OK;
		break;
	}

	fclose(fd);
	X509_free(cert);

	return r;
}


/*
 * Print information about the SSL/TLS certificate.
 */
void
print_cert(X509 * cert, unsigned char *md, unsigned int *mdlen)
{
	unsigned int i;
	char *c;

	c = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
	printf("Server certificate subject: %s\n", c);
	xfree(c);

	c = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
	printf("Server certificate issuer: %s\n", c);
	xfree(c);

	printf("Server key fingerprint: ");
	for (i = 0; i < *mdlen; i++)
		printf(i != *mdlen - 1 ? "%02X:" : "%02X\n", md[i]);
}


/*
 * Write the SSL/TLS certificate after asking the user to accept/reject it.
 */
int
write_cert(X509 * cert)
{
	FILE *fd;
	char c, buf[LINE_MAX];

	do {
		printf("(R)eject, accept (t)emporarily or "
		    "accept (p)ermanently? ");
		fgets(buf, LINE_MAX, stdin);
		c = tolower(*buf);
	} while (c != 'r' && c != 't' && c != 'p');

	if (c == 'r')
		return SSL_CERT_ACTION_REJECT;
	else if (c == 't')
		return SSL_CERT_ACTION_ACCEPT;

	create_file(PATHNAME_CERT_FILE, S_IRUSR | S_IWUSR);

	fd = fopen(PATHNAME_CERT_FILE, "a");
	if (fd == NULL)
		return SSL_CERT_ACTION_REJECT;

	PEM_write_X509(fd, cert);

	fclose(fd);

	return SSL_CERT_ACTION_ACCEPT;
}


/*
 * Ask user to proceed, while a fingerprint mismatch in the SSL/TLS
 * certificate was found.
 */
int
mismatch_cert(void)
{
	char c, buf[LINE_MAX];

	do {
		printf("WARNING: SSL/TLS certificate fingerprint mismatch.\n"
		    "Proceed with the connection (y/n)? ");
		fgets(buf, LINE_MAX, stdin);
		c = tolower(*buf);
	} while (c != 'y' && c != 'n');

	if (c == 'y')
		return SSL_CERT_ACTION_CONTINUE;
	else
		return SSL_CERT_ACTION_ABORT;
}
#endif				/* SSL_TLS */
