#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "config.h"
#include "imapfilter.h"
#include "pathnames.h"


extern char *home;


/*
 * Create imapfilter's home directory.
 */
int
create_homedir(void)
{
	if (home != NULL)
		if (chdir(home))
			error("could not change directory; %s\n",
			    strerror(errno));

	if (!exists_dir(PATHNAME_HOME_DIR)) {
		if (mkdir(PATHNAME_HOME_DIR, S_IRUSR | S_IWUSR | S_IXUSR))
			error("could not create directory %s; %s\n",
			    PATHNAME_HOME_DIR, strerror(errno));
	}
#ifdef CHECK_PERMISSIONS
	else {
		check_dir_perms(PATHNAME_HOME_DIR, S_IRUSR | S_IWUSR | S_IXUSR);
	}
#endif

	return 0;
}


/*
 * Check if a file exists.
 */
int
exists_file(char *fname)
{
	struct stat fs;

	if (access(fname, F_OK))
		return 0;

	stat(fname, &fs);
	if (!S_ISREG(fs.st_mode)) {
		error("file %s not a regular file\n", fname);
		return ERROR_FILE_OPEN;
	}
	return 1;
}


/*
 * Check if a directory exists.
 */
int
exists_dir(char *dname)
{
	struct stat ds;

	if (access(dname, F_OK))
		return 0;

	stat(dname, &ds);
	if (!S_ISDIR(ds.st_mode)) {
		error("file %s not a directory\n", dname);
		return ERROR_FILE_OPEN;
	}
	return 1;
}


/*
 * Create a file with the specified permissions.
 */
int
create_file(char *fname, mode_t mode)
{
	int fd;

	fd = 0;

	if (!exists_file(fname)) {
		fd = open(fname, O_CREAT | O_WRONLY | O_TRUNC, mode);
		if (fd == -1) {
			error("could not create file %s; %s\n", fname,
			    strerror(errno));
			return ERROR_FILE_OPEN;
		}
		close(fd);
	}
	return 0;
}


#ifdef CHECK_PERMISSIONS
/*
 * Check the permissions of a file.
 */
int
check_file_perms(char *fname, mode_t mode)
{
	struct stat fs;

	if (stat(fname, &fs)) {
		error("getting file %s status; %s\n", fname,
		    strerror(errno));
		return ERROR_TRIVIAL;
	}
	if (!S_ISREG(fs.st_mode)) {
		error("file %s not a regular file\n", fname);
		return ERROR_TRIVIAL;
	}
	if ((fs.st_mode & 00777) != mode) {
		error("warning: improper file %s permissions\n", fname);
		error("warning: file's mode should be %o not %o\n", mode,
		    fs.st_mode & 00777);
		return ERROR_TRIVIAL;
	}
	return 0;
}


/*
 * Check the permissions of a directory.
 */
int
check_dir_perms(char *dname, mode_t mode)
{
	struct stat ds;

	if (stat(dname, &ds)) {
		error("getting file %s status; %s\n", dname,
		    strerror(errno));
		return ERROR_TRIVIAL;
	}
	if (!S_ISDIR(ds.st_mode)) {
		error("file %s not a directory\n", dname);
		return ERROR_TRIVIAL;
	}
	if ((ds.st_mode & 00777) != mode) {
		error("warning: improper dir %s permissions\n", dname);
		error("warning: dir's mode should be %o not %o\n", mode,
		    ds.st_mode & 00777);
		return ERROR_TRIVIAL;
	}
	return 0;
}
#endif				/* CHECK_PERMISSIONS */
